/* Name: main.c
 * Project: Tektronix 4957 Tablet USB Adapter
 * Author: Matthias Schldgen DC7UR
 * Creation Date: 2009-05-11
 * Tabsize: 4
 * Copyright: (c) 2009 Matthias Schldgen
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: main.c 692 2008-11-07 15:07:40Z cs $
 */
/*
Based on V-USB , a software-only library to provide USB functionality with AVR's . 
get more info about V-USB at http://www.obdev.at

This project runs on the ATTiny 45/85 ( code size is too large for the 25 ) and should 

1. initialize the Tablet to movement streaming, binary mode. binary mode uses considerably less 
bytes than ASCII mode and saves some power on the AVR although the data format is a bit obscure
2. provide an absolute pointer device to USB host 
3. shamelessly grab Tektronix Vendor ID (0x0699 )  and Device ID ( 4957 ) hehehe :P 
4. we use our interupt driven serial receiver code to get tablet data . 
As sending to the tablet is only done in init we save the effort of
implementing the serial TX code in interrupt. ( as i did on the Arduino board to debug the tablet)  
Using the pinchange interrupt ( PC_INT0 ) shouldn't interfere with the V-USB code as it has lower priority.
Baudrate is 9600 which gives us a lot of time to do the work. 
The major challenge is the software serial receiver routine which has to be interrupt driven for fast response 

The Tektronix is a quality inductive tablet with four buttons and a maximum resolution 
of approx. 24000 dots on its 12" * 12" ( = 30cm * 30cm ) surface.
Originally it is supplied by -12v / +12V from the special interface card or supply. I use a 
DC-DC Converter to generate the +/-12Volts from a single 5 Volt supply. 
Unfortunately the current is above the allowed 500mA for USB and therefore it needs its own power. 

The USB I/F is buspowered and draws less than 20mA , including the MAX232 . Note that there's a bug 
in V-USB which requires you to use PB2 and PB0 for USB on the ATTiny 45/85, 
other pins than PB0 for D- do not work. (and PB2 is INT0 which you have to use anyway)   

This ATTiny runs from an ext. 16Mhz clock as i had no luck ever to calibrate the d**ed things :P . 
Connect the oscillator to PB3 . 

Note that this project heavily relies on the ATTiny 45/85 hardware within the IRQ routines . 
For fastest program execution times the SFR's are directly programmed. A future release would
probably use ASM for the IRQ driven serial port. More about the communication with the tablet is 
explained below. 
Comm. parameters are 9600,O,7,1 , which means we have to pull in the parity macro

References : 
binary mode :
xf86Tek4957.c - Tektronix Tablet driver for X (c) 2002 by Olivier DANET 
ascii mode :
MAPFIX.BAS : make your own maps with the help of WA1VJH and W7KKE 

Serial to USB converter on Arduino Duemilanove for testing IRQ driven serial routines. (ASM project )

Bitbanging in C is not trivial , i'm much more familiar with it in ASM. I learned a lot from the 
'i2c_tiny_usb' source code - thanks to Till Harbaum  ! www.harbaum.org/till/i2c_tiny_usb/

built with AVR Studio 4 and WinAVR

*/
#include <stdint.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>  /* for sei() */
#include <util/delay.h>     /* for _delay_ms() */
#include <util/parity.h>     /* parity for tektronix */

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include "usbdrv.h"
#include "oddebug.h"        /* This is also an example for using debug macros */

// useful macros for for extracting bytes from ints
#define LO(x) ( (uchar) ((x) & 0x00ff))
#define HI(x) ( (uchar) ((x) >> 8) & 0x00ff)) 
/* software serial port */
/* taken from the assembler source of my arduino project */
/* timer periods for 9600 baud, 16mhz and a prescaler of 32 */
/* resolution is 2 uSec per timer tick					*/ 
#define period 256-52			/* time from bit to bit = 104 uSecs*/
#define startperiod 256-74 		/* time from start of start bit to middle of LSB */ 

/* usecs periods for 9600 baud used in the serial transmit routine */
#define bitrate 104 

/* hardware defs for serial port */
#define BIT_RXD 4
#define BIT_TXD 1

/* tablet comm char */
#define ESC 0x1b


#ifndef FALSE
#define FALSE 0x0
#endif

/* Tektronix data string is 
 VV BB LX MX HX LY MY HY 
 where 
 VV = sync and data validity: 44 = valid , 45 = invalid
 BB = buttons in bits 00004321 
 LX MX HX = X value in 6bit bytes 
 LY MY HY = Y value in 6bit bytes 
*/

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */
/* declare a digitzer device generated with HID Descriptor tool */
/* 4 buttons (reports only 3 for compatibility), absolute pointer , normally generates X/Y valus of max. 15 bits    
 * this represents an absolute pointer device with a resolution of 0.1 mm (2972 dots on a 29,7 cm tablet)
 * to the OS . 
*/
PROGMEM char usbHidReportDescriptor[60] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x02,                    // USAGE (Mouse)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x09, 0x01,                    //   USAGE (Pointer)
    0xa1, 0x00,                    //   COLLECTION (Physical)
    0x05, 0x09,                    //     USAGE_PAGE (Button)
    0x19, 0x01,                    //     USAGE_MINIMUM (Button 1)
    0x29, 0x03,                    //     USAGE_MAXIMUM (Button 3)
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x75, 0x01,                    //     REPORT_SIZE (1)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0x95, 0x01,                    //     REPORT_COUNT (1)
    0x75, 0x05,                    //     REPORT_SIZE (5)
    0x81, 0x03,                    //     INPUT (Cnst,Var,Abs)
    0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x35, 0x00,                    //     PHYSICAL_MINIMUM (0)
    0x46, 0x9d, 0x0b,              //     PHYSICAL_MAXIMUM (2973)
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x26, 0x9d, 0x0b,              //     LOGICAL_MAXIMUM (2973)
    0x65, 0x11,                    //     UNIT (SI Lin:Distance)
    0x55, 0x0e,                    //     UNIT_EXPONENT (-2)
    0x75, 0x10,                    //     REPORT_SIZE (16)
    0x95, 0x02,                    //     REPORT_COUNT (2)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0xc0,                          //   END_COLLECTION
    0xc0                           // END_COLLECTION
};
// PROGMEM int usbDescriptorStringDevice[] = { USB_STRING_DESCRIPTOR_HEADER(14),
// 'T','e','k','t','r','o','n','i','x',' ','4','9','5','7'
// };
/* here is the Init string for the tablet 
 * a colon tells the sendroutine to pause for a few dozen mSecs
 * reset , binary output  
 * resolution 2972 = 0.1 mm   
 * report speed normal
 * topleft is zero 
 * emit when movement 
 * continous output
 * activate configuration
 * report resolution
 */
PROGMEM char TabletInit[33] = {
	ESC,'Z',':',
	ESC,'C','1',':',
	ESC,'R','2',':',
	ESC,'F','3',':',
	ESC,'I','0','0','1',':',
	ESC,'M','0',':',
	ESC,'x',':',':',
	ESC,'a',':',':'
};

typedef struct{
 	uchar		buttonMask;
	uint16_t	dx ;
	uint16_t	dy ;
//    char    dWheel;
}report_t;

static 	report_t reportBuffer;
static 	uchar    idleRate;   /* repeat rate for keyboards, never used for mice */
/* softserial IRQ values */
volatile uint8_t	offset;		/* bit shifter */
volatile uint8_t	val;		/* receive value */
/* buffer properties */
static 	uint8_t 	rxbuf[16];
static 	uint8_t 	rxpos;
/* cursor coordinates */ 
static 	uint16_t	newposy, newposx;
static 	uint8_t	oldpos4, oldpos7;

/* Let us once more implement the serial receiver routine in interrupt, this time in C.
PCINT0 is fired by the startbit of incoming serial and then prepares variables and starts the timer 
- after that it blocks itself . It is freed again by the completion of the timer routine. 
never touch the global irq flag 'cause the USB routines need it */
ISR(PCINT0_vect,ISR_NOBLOCK)
{
  if (bit_is_clear(PINB,BIT_RXD))		/* only a falling edge will trigger */
    {
      /* Start timer 1 with a prescaler of 32 */
      TCNT1 = startperiod;
      TCCR1 |= _BV(CS12) | _BV(CS11);	/* start with prescaler / 32  */
	  GIMSK &= ~(_BV(PCIE));	/* block myself */
 	  GIFR |= _BV(PCIF); 		/* clr pendings on me */
	  TIFR	|= _BV(TOV1);		/* clr pending on timer */
	  TIMSK |= _BV(TOIE1);		/* enable Timer 1 irq */
	  offset = 0;				/* shifter init */
	  val = 0;					/* value init */
    }
	return;
};
/* the timer routine takes care of incoming bits from software serial 
 * it is initialized by the PC_INT0 Routine . the timer IRQ now sets the timer to the 
 * real bit rate and samples the incoming data line in the middle of the bit 
 * keep in mind that RS232 sends the LSB first in contrary to e.g. I2C. 
 * 
 * the assembled value is in 'val' 
 */
ISR(TIM1_OVF_vect,ISR_NOBLOCK)
{
 	TCNT1 = period ;		/* restart timer in any case to maintain best timing */
	if ( offset > 7 ) {			/* swallow stop bit */
	  	TIMSK &= ~(_BV(TOIE1));	/* block myself */
 		GIFR |= _BV(PCIF); 		/* clr pendings on PC_INT*/
 		PCMSK |= _BV(BIT_RXD); 	/* prepare pinchange irq */	
  		TCCR1 = 0;				/* stop timer */
		GIMSK |= _BV(PCIE);		/* enable PC INT0 we don't expect a Pinchange between stop bit and idle */
		offset += 1;			/* at the end of a completed byte we should have offset == 9 */
		return;					/* out of here */
	}
	val >>= 1; 					/* shift one bit to the right - LSB comes first */ 
	if (bit_is_set(PINB,BIT_RXD)) {
		val |= 0x80 ; /* set the highest data bit */
	}
	offset += 1;		/* outerloop counter */
	return;	
};

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
usbRequest_t    *rq = (void *)data;

    /* The following requests are never used. But since they are required by
     * the specification, we implement them in this example.
     */
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
 //       DBG1(0x50, &rq->bRequest, 1);   /* debug output: print our request */
        if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
            /* we only have one report type, so don't look at wValue */
            usbMsgPtr = (void *)&reportBuffer;
            return sizeof(reportBuffer);
        }else if(rq->bRequest == USBRQ_HID_GET_IDLE){
            usbMsgPtr = &idleRate;
            return 1;
        }else if(rq->bRequest == USBRQ_HID_SET_IDLE){
            idleRate = rq->wValue.bytes[1];
        }
    }else{
        /* no vendor specific requests implemented */
    }
    return 0;   /* default for not implemented requests: return no data back to host */
};
/* serial output to tablet . this is more or less only called during init 
so we save the effort of making it irq controlled for now. */
void serout (const uchar s)
{
uchar mask ;
/* start bit */
    PORTB &= ~(_BV(BIT_TXD));	/* low for start bit  */
	_delay_us(bitrate);
 	for (mask = 0x01; mask; mask <<= 1) {
    	if (s & mask) { // choose bit
    		PORTB |= _BV(BIT_TXD);
     	} else {
			PORTB &= ~(_BV(BIT_TXD));
   		}
	_delay_us(bitrate);
 	}
    PORTB |= _BV(BIT_TXD);			/* stop bit */	
	_delay_us(bitrate * 2);			/* send  empty period */
};

/* fast serial output for debug only */
void debugout(const uchar s)
{
uchar mask ;
/* start bit */
    PORTB &= ~(_BV(BIT_TXD));	/* low for start bit  */
	_delay_us(8);
 	for (mask = 0x01; mask; mask <<= 1) {
    	if (s & mask) { // choose bit
    		PORTB |= _BV(BIT_TXD);
     	} else {
			PORTB &= ~(_BV(BIT_TXD));
   		}
	_delay_us(2);
 	}
    PORTB |= _BV(BIT_TXD);			/* stop bit */	
};
/* get the init string from ROM and stream it to the tablet */
/* uses parity function we could hardcode the parity but we need the function elsewhere anyway*/ 
void inittablet(const char *s) {
uchar c;
	for (c = pgm_read_byte(s); c ; ++s, c = pgm_read_byte(s) ) {
		if ( c == ':' ) {		/* a colon means wait time */
			_delay_ms(60);				/* tablet execution time */
		} else {
			if (!(parity_even_bit(c))) c |= 0x80;	/* add odd parity */
			serout(c);				/* send the char */
		}
	}
	return;
};
/* bitbang surveillance 
 * checks for a complete byte in val 
 * 
 */
/* central tablet evaluation routine 
 * first check if we have new data then put the received byte into buffer.
 * sync byte is the leading 0x44 or 0x45 byte , followed by 7 data bytes - only the sync has bit 6 set. 
 */
static void updateBuffer(void)
{
	if ( offset > 8 ) {		/* the marker for our irq routine completion */
/* use parity for simple error checking. btw, parity_even bit() returns true if parity is odd !*/ 
		if (parity_even_bit(val)) {
			if (val & 0x40)
			{
				rxpos =  1;    /* sync byte is the only one with this bit set */
				}
			val &= 0x3f ; 				/* strip bit 6 & 7 */ 
			rxbuf[rxpos] = val;			/* put in receiver buffer */
			rxpos += 1;					/* next pos in receiverbuffer */
			offset = 0;					/* reset */
			if (rxpos > 7) {			/* ok, we should have a complete data set */
			if  ((rxbuf[1] & 0x01) | ((rxbuf[4]==0) & (oldpos4 !=0) ) | ((rxbuf[7]==0) & (oldpos7 !=0))) 
			{
/* valid data ? my tablet strays a few low coordinates in certain regions. I mask out values */
					reportBuffer.dx = 0x7fff;		/* invalid data helps to put focus back to mouse */ 
					reportBuffer.dy = 0x7fff;
					}
				else {
//				debugout(rxbuf[4]);
/* reassemble the Tek format to something you can work with */
					reportBuffer.buttonMask = rxbuf[2] & 0x07 ;
					newposx = rxbuf[3] + (64 * rxbuf[4]); //  + (4096 * rxbuf[5]);	
					newposy = rxbuf[6] + (64 * rxbuf[7]); //  + (4096 * rxbuf[8]);
					reportBuffer.dx =newposx; // ((newposx + oldposx) / 2) & 0x0fff ; // ((LO(newposx) << 8) | (HI(newposx)) ;
					reportBuffer.dy =newposy; //((newposy + oldposy) / 2) & 0x0fff ; //((LO(newposy) << 8) | (HI(newposy)) ;
				}
			oldpos4 =  rxbuf[4]; 
			oldpos7 =  rxbuf[7]; 
			rxpos = 1;
		 }
		}
		}
	return;
}

/* ------------------------------------------------------------------------- */
/* MAIN routine 
 * this looks dangerously close to the mouse HID example from obdev.at and is in fact stolen from it 
 * added hardware inits and changed to update the buffers faster. 
 */
int main(void)
{
    	wdt_enable(WDTO_8S); /* the 8 seconds wd time is not possible on all AVR's */
//	wdt_disable();
    	/* Even if you don't use the watchdog, turn it off here. On newer devices,
     	* the status of the watchdog (on/off, period) is PRESERVED OVER RESET!
     	*/
    DBG1(0x00, 0, 0);       /* debug output: main starts */
    /* RESET status: all port bits are inputs without pull-up.
     * That's the way we need D+ and D-. */
	/* hardware init for the Tektronix converter */
    DDRB |= _BV(BIT_TXD);   /* output for serial to tablet */
    PORTB |= _BV(BIT_RXD);  /* enable pullup on RXD pin */
    PORTB |= _BV(BIT_TXD);	/* set TXD hi (inactive ) */
	TCCR1 = 0;				/* preset timer register to defined state */
 	GIFR |= _BV(PCIF); 		/* clr pendings */
 	PCMSK = 0x0 | _BV(BIT_RXD); 	/* prepare pinchange irq */	
/* now initialize software components */
	odDebugInit();
	usbInit();
	usbDeviceDisconnect();  	/* enforce re-enumeration, do this while interrupts are disabled! */
	_delay_ms(300);
	usbDeviceConnect();
/* enable serial irq routines */
	GIFR |= _BV(PCIF); 			/* clr pendings */
	GIMSK |= _BV(PCIE); 		/* enable PC_INT */
	sei();						/* globally enable IRQ's */
	inittablet(TabletInit);		/* the tablet sends data after init therefore init after irqs */ 	
	_delay_ms(500);
	rxpos = 1;					/* reset input buffer pointer once */ 
    DBG1(0x01, 0, 0);       	/* debug output: main loop starts */
    for(;;){                	/* main event loop */
        DBG1(0x02, 0, 0);   	/* debug output: main loop iterates */
       wdt_reset();
        usbPoll();
        updateBuffer();			/* main tablet routine */
        if(usbInterruptIsReady()){
            /* called after every poll of the interrupt endpoint */
            	DBG1(0x03, 0, 0);   /* debug output: interrupt report prepared */
            	usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
        }
    }
    return 0;
}

/* ------------------------------------------------------------------------- */
